#include "FlashMe.h"

UCHAR			lpBlock[NAND_BLOCK_SIZE];
WCHAR			lpBuffer[MAX_PATH];
USHORT			crc16;

int WINAPI WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nCmdShow )
{
    INITCOMMONCONTROLSEX	iccex;
	WNDCLASS				wc;
	MSG						msg;
	HWND					hWndMain;

	iccex.dwSize = sizeof(iccex);
	iccex.dwICC	 = ICC_PROGRESS_CLASS;
	InitCommonControlsEx( &iccex );

	wc.style			= CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc		= (WNDPROC) WndProc;
    wc.cbClsExtra		= 0;
    wc.cbWndExtra		= DLGWINDOWEXTRA;
    wc.hInstance		= hInstance;
    wc.hIcon			= LoadIcon(hInstance, MAKEINTRESOURCE(IDI_MAIN));
    wc.hCursor			= 0;
    wc.hbrBackground	= (HBRUSH)(COLOR_BTNFACE+1);
    wc.lpszMenuName		= 0;
    wc.lpszClassName	= TEXT("MainClass32");
	RegisterClass( &wc );
	
	hWndMain = CreateDialogParam( hInstance, MAKEINTRESOURCE(IDD_MAIN), 0, (DLGPROC)WndProc, (LPARAM)lpCmdLine );
	ShowWindow( hWndMain, SW_SHOW );
	SetWindowPos(hWndMain, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
	UpdateWindow( hWndMain );
	
	while( GetMessage( &msg, 0, 0, 0 ) )
	{
		TranslateMessage( &msg );
		DispatchMessage( &msg );
	}
	return (int)msg.wParam;
}


LRESULT CALLBACK WndProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	switch( uMsg )
	{
	case WM_INITDIALOG:
		if( !CreateThread( 0, 0, NAND_WorkerThread, hWnd, 0, 0 ) )
		{
			SetDlgItemText( hWnd, IDC_STATUS, TEXT("error instantiating worker thread ...") );
		}
		else
		{
			SetDlgItemText( hWnd, IDC_STATUS, TEXT("ready ...") );
		}
		break;	
	
	case WM_CLOSE:
		DestroyWindow( hWnd );
		break;

	case WM_DESTROY:
		PostQuitMessage( 0 );
		break;

	default:
		return DefWindowProc( hWnd, uMsg, wParam, lParam );
	}
	
	return 0;
}


ULONG WINAPI NAND_WorkerThread( LPVOID hWnd )
{
	HINSTANCE		hModule;
	WIN32_FIND_DATA	FindFileData;
	HANDLE			hFindFile, hFile;
	BOOL			bFindFile;
	ULONG			lpBytesRead, HeaderSize, WriteAddress, EndAddress, lpPhysicalAddress;
	LPVOID			lpVirtualAddress;
	FLASHHEADER		FlashHeader;
	UCHAR			i;
	UCHAR			SpareData[16];

	/* init interface */
	hModule = NAND_GetInterface();
	if( !hModule )
	{
		SetDlgItemText( (HWND)hWnd, IDC_STATUS, TEXT("error instantiating interface ...") );
		ExitThread( 0 );
	}
	
	/* init memory and nand */
	lpVirtualAddress = NAND_AllocPhysMem( NAND_BLOCK_SIZE, PAGE_NOCACHE | PAGE_READWRITE, 0, 0, &lpPhysicalAddress );
	NAND_Init( lpVirtualAddress, lpPhysicalAddress );

	/* find first bin file */
	hFindFile = FindFirstFile( TEXT("\\USBDisk\\*.bin"), &FindFileData );
	if( hFindFile == INVALID_HANDLE_VALUE )
	{
		SetDlgItemText( (HWND)hWnd, IDC_STATUS, TEXT("no flashable file found ...") );
		ExitThread( 0 );
	}
	bFindFile = TRUE;


	/* find file loop */
	while( bFindFile != FALSE )
	{
		SetDlgItemText( (HWND)hWnd, IDC_FLASHFILE, FindFileData.cFileName );
		
		lstrcpy( lpBuffer, TEXT("\\USBDisk\\") );
		lstrcat( lpBuffer, FindFileData.cFileName );
	
		/* open file for reading */
		hFile = CreateFile( lpBuffer, GENERIC_READ, 0, 0, OPEN_EXISTING, 0, 0 );
		if( hFile == INVALID_HANDLE_VALUE )
		{
			SetDlgItemText( (HWND)hWnd, IDC_STATUS, TEXT("error opening file ...") );
			CloseHandle( hFindFile );
			ExitThread( 0 );
		}

		/* read header */
		HeaderSize = 0;	
		ReadFile( hFile, &FlashHeader.version, sizeof(FlashHeader.version), &lpBytesRead, 0 );
		HeaderSize += sizeof(FlashHeader.version);
		ReadFile( hFile, &FlashHeader.crc16, sizeof(FlashHeader.crc16), &lpBytesRead, 0 );
		HeaderSize += sizeof(FlashHeader.crc16);
		ReadFile( hFile, &FlashHeader.flags, sizeof(FlashHeader.flags), &lpBytesRead, 0 );
		HeaderSize += sizeof(FlashHeader.flags);
		ReadFile( hFile, &FlashHeader.start, sizeof(FlashHeader.start), &lpBytesRead, 0 );
		HeaderSize += sizeof(FlashHeader.start);
		ReadFile( hFile, &FlashHeader.size, sizeof(FlashHeader.size), &lpBytesRead, 0 );
		HeaderSize += sizeof(FlashHeader.size);
		ReadFile( hFile, &FlashHeader.reserved, sizeof(FlashHeader.reserved), &lpBytesRead, 0 );
		HeaderSize += sizeof(FlashHeader.reserved);
		if( SetFilePointer(hFile, 0, 0, FILE_CURRENT) != HeaderSize )
		{
			SetDlgItemText( (HWND)hWnd, IDC_STATUS, TEXT("error reading header ...") );
			break;
		}
		
		/* init gui stuff */
		wsprintf( lpBuffer, TEXT("0x%04X"), FlashHeader.crc16 );
		SetDlgItemText( (HWND)hWnd, IDC_FLASHCRC, lpBuffer );
		wsprintf( lpBuffer, TEXT("0x%08X"), FlashHeader.start );
		SetDlgItemText( (HWND)hWnd, IDC_FLASHSTART, lpBuffer );
		wsprintf( lpBuffer, TEXT("0x%08X"), FlashHeader.size );
		SetDlgItemText( (HWND)hWnd, IDC_FLASHSIZE, lpBuffer );
		wsprintf( lpBuffer, TEXT("0x%02X"), FlashHeader.flags );
		SetDlgItemText( (HWND)hWnd, IDC_FLASHECC, lpBuffer );
		SendDlgItemMessage( hWnd, IDC_PROGRESS, PBM_SETRANGE32, 0, (FlashHeader.size / NAND_PAGE_SIZE) );
		SendDlgItemMessage( hWnd, IDC_PROGRESS, PBM_SETSTEP, 1, 0 );
		SendDlgItemMessage( hWnd, IDC_PROGRESS, PBM_SETPOS, 0, 0 );

		/* validate checksum */
		SetDlgItemText( (HWND)hWnd, IDC_STATUS, TEXT("validating checksum ...") );
		WriteAddress = 0;
		EndAddress	 = FlashHeader.size;
		crc16		 = 0xFFFF;
		while( WriteAddress < EndAddress )
		{
			ReadFile( hFile, lpBlock, NAND_BLOCK_SIZE, &lpBytesRead, 0 );
			NAND_GetBlockCRC( lpBlock );
			WriteAddress += NAND_BLOCK_SIZE;
		}
		if( crc16 != FlashHeader.crc16 )
		{
			SetDlgItemText( (HWND)hWnd, IDC_STATUS, TEXT("error validating checksum ...") );
			break;
		}
		
		/* flash file */
		SetFilePointer( hFile, HeaderSize, 0, FILE_BEGIN );
		SetDlgItemText( (HWND)hWnd, IDC_STATUS, TEXT("flashing ...") );
		WriteAddress = FlashHeader.start;
		EndAddress	 = WriteAddress + FlashHeader.size;
		while( WriteAddress < EndAddress )
		{
			if( !NAND_EraseBlock( WriteAddress ) )
			{
				WriteAddress += NAND_BLOCK_SIZE;
			}
			else
			{
				for( i=0; i<NAND_NROF_PAGES; i++ )
				{
					ReadFile( hFile, lpVirtualAddress, NAND_PAGE_SIZE, &lpBytesRead, 0 );
					NAND_WritePage( WriteAddress, lpVirtualAddress, lpPhysicalAddress );
						
					switch( FlashHeader.flags )
					{
					case 1:			
						// linux_ecc only
						memset( SpareData, 0xFF, 16 );
						linux_old_calc_ecc( (UCHAR *)lpVirtualAddress, &SpareData[0] );
						linux_old_calc_ecc( ((UCHAR *)lpVirtualAddress)+256, &SpareData[3] );
						SpareData[6] = SpareData[4];
						SpareData[7] = SpareData[5];
						SpareData[4] = 0xFF;
						SpareData[5] = 0xFF;
						NAND_WriteSpareData( WriteAddress, &SpareData );
						break;
					case 2:
						// wince_ecc and linux_ecc
						NAND_ReadSpareData( WriteAddress, &SpareData );
						linux_old_calc_ecc( (UCHAR *)lpVirtualAddress, &SpareData[0] );
						linux_old_calc_ecc( ((UCHAR *)lpVirtualAddress)+256, &SpareData[3] );
						SpareData[6] = SpareData[4];
						SpareData[7] = SpareData[5];
						SpareData[4] = 0xFF;
						SpareData[5] = 0xFF;
						NAND_WriteSpareData( WriteAddress, &SpareData );
					default:
						// wince_ecc only
						break;
					}
					SendDlgItemMessage( hWnd, IDC_PROGRESS, PBM_STEPIT, 0, 0 );
					WriteAddress += NAND_PAGE_SIZE;
				}
			}
		}
		/*
		WriteAddress = EndAddress;
		EndAddress = 0x03C00000;
		while( WriteAddress < EndAddress )
		{
			NAND_EraseBlock( WriteAddress );
			WriteAddress += NAND_BLOCK_SIZE;
		}
		*/

		bFindFile = FindNextFile( hFindFile, &FindFileData );;
		SetDlgItemText( (HWND)hWnd, IDC_STATUS, TEXT("ready ...") );
	}
	
	SetDlgItemText( (HWND)hWnd, IDC_STATUS, TEXT("finished ...") );
	CloseHandle( hFindFile );
	CloseHandle( hFile );
	ExitThread( 0 );

	return 0;
}


HMODULE NAND_GetInterface()
{
	HINSTANCE	hModule;
	HINSTANCE	hInterface;
	
	hModule				= LoadLibrary( TEXT("\\USBDisk\\flashlib.dll") );
	hInterface			= (HINSTANCE)GetProcAddress( hModule, TEXT("TmGetUncachedDSPMemory") );				// 1000AA38

	NAND_AllocPhysMem	= (PROTO_ALLOCMEM)  (hInterface + ((0x10010DA4 - 0x1000AA38) / 4) );				// 10010DA4
	NAND_FreePhysMem	= (PROTO_FREEMEM)   (hInterface + ((0x10010D84 - 0x1000AA38) / 4) );				// 10010D84
	NAND_Init			= (PROTO_NANDINIT)  (hInterface + ((0x100139E0 - 0x1000AA38) / 4) );				// 100139E0
	NAND_IsBlockBad		= (PROTO_BADBLOCK)  (hInterface + ((0x10012AC0 - 0x1000AA38) / 4) );				// 10012AC0
	NAND_EraseBlock		= (PROTO_ERASEBLOCK)(hInterface + ((0x10013194 - 0x1000AA38) / 4) );				// 10013194
	NAND_ReadPage		= (PROTO_READPAGE)  (hInterface + ((0x10012934 - 0x1000AA38) / 4) );				// 10012934
	NAND_WritePage		= (PROTO_WRITEPAGE) (hInterface + ((0x1001311C - 0x1000AA38) / 4) );				// 1001311C
	NAND_ReadSpareData	= (PROTO_READSPARE) (hInterface + ((0x10012794 - 0x1000AA38) / 4) );				// 10012794
	NAND_WriteSpareData	= (PROTO_WRITESPARE)(hInterface + ((0x10012D80 - 0x1000AA38) / 4) );				// 10012D80
	
	return hModule;
}


USHORT NAND_GetBlockCRC( UCHAR *data )		// CRC-CCITT (0xFFFF)
{
	USHORT i, crc_new;
	UCHAR x;

	for (i=0; i < NAND_BLOCK_SIZE; i++)
	{
		x = data[i];
		crc_new = (unsigned char)(crc16 >> 8) | (crc16 << 8);
		crc_new ^= x;
		crc_new ^= (unsigned char)(crc_new & 0xff) >> 4;
		crc_new ^= crc_new << 12;
		crc_new ^= (crc_new & 0xff) << 5;
		crc16 = crc_new;
	}
	return crc16;
} 


